#include <array>

#include "PluginStatusCodes.hpp"

#include "ExampleTracker.hpp"

/**
 * Initializes tracker
 */
ExampleTracker::ExampleTracker(                      //
    const varjo_PluginUtilityAPI* pluginUtilityAPI,  //
    varjo_PluginUtilitySession pluginUtilitySession)
    : m_pluginUtilityAPI(pluginUtilityAPI)
    , m_pluginUtilitySession(pluginUtilitySession)
{
    if (m_pluginUtilityAPI && m_pluginUtilitySession) {
        // Plugin author is free to define configuration settings
        // necessary for the tracker plugin
        m_pluginUtilityAPI->SetConfigurationString(  //
            m_pluginUtilitySession, "server_address", "localhost");
        m_pluginUtilityAPI->SetConfigurationString(  //
            m_pluginUtilitySession, "headset_id", "0");
    }
}

/**
 * Starts tracker processing
 */
varjo_PluginStatus ExampleTracker::start()
{
    if (!m_pluginUtilityAPI || !m_pluginUtilitySession) {  //
        return varjo_PluginStatus_NullPointer;
    }

    // Tracker can query system time to synchronize with Varjo software
    // stack, (e.g. to bring tracker time into the same epoch as the
    // requests for the HMD poses)
    m_startTime = m_pluginUtilityAPI->GetCurrentSystemTime(m_pluginUtilitySession);

    m_isRunning = true;
    m_processingThread = std::thread(&ExampleTracker::trackerMainLoop, this);
    if (!m_processingThread.joinable()) {
        m_isRunning = false;
        return varjo_PluginStatus_TrackerInitializationError;
    }

    return varjo_PluginStatus_OK;
}

/**
 * Stops tracker processing
 */
varjo_PluginStatus ExampleTracker::stop()  //
{
    if (!m_isRunning)  //
        return varjo_PluginStatus_OK;

    m_isRunning = false;
    m_processingThread.join();

    return varjo_PluginStatus_OK;
}

/**
 * Returns tracked HMD pose
 */
varjo_PluginStatus ExampleTracker::getHMDPose(            //
    varjo_Nanoseconds timeStamp, varjo_DevicePose* pose)  //
{
    if (!pose)  //
        return varjo_PluginStatus_NullPointer;

    if (!m_isRunning)  //
        return varjo_PluginStatus_TrackingNotInitialized;

    pose->poseFlags = varjo_PoseFlags_Ok;
    pose->frameNo = m_frameNumber++;
    pose->timeStamp = timeStamp;

    pose->position = {0., 0., 0.};
    pose->rotation = {1., 0., 0., 0.};
    pose->acceleration = {0., 0., 0.};
    pose->velocity = {0., 0., 0.};
    pose->angularVelocity = {0., 0., 0.};
    pose->confidence = 1.;

    return varjo_PluginStatus_OK;
}

/**
 * Main processing loop of the tracker
 */
void ExampleTracker::trackerMainLoop()  //
{
    std::array<varjo_ImuSample, 64> imuSamples;

    while (m_isRunning) {
        // Using ReadIMUSamples methods of utility API Tracker can query IMU data:
        const int32_t readSamples = m_pluginUtilityAPI->ReadIMUSamples(  //
            m_pluginUtilitySession, imuSamples.data(), (int32_t)imuSamples.size());
        if (readSamples > 0) {
            // Process IMU samples
        }

        // Tracker processing happens here

        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }
}
