// Copyright 2019 Varjo Technologies Oy. All rights reserved.

#ifndef VARJO_PLUGIN_H
#define VARJO_PLUGIN_H

#include "Varjo_export.h"
#include "Varjo_types.h"
#include "Varjo_version.h"

#if defined __cplusplus
extern "C" {
#endif  // __cplusplus

#if defined(_WIN32)
#define VARJO_FNTABLE_CALLTYPE __cdecl
#else
#define VARJO_FNTABLE_CALLTYPE
#endif

// ----------------------------------------------------------------------------
// API types

/**
 * Bit-field type determining type of plugin API
 */
typedef uint64_t varjo_PluginType;
static const varjo_PluginType varjo_PluginType_HMD_Tracker = 0x01;  //!< Flag indicating this plugin provides HMD tracking functionality
// static const varjo_PluginType varjo_PluginType_InputDevice = 0x02;

// ----------------------------------------------------------------------------
// Plugin API

/**
 * Provides plugin types supported by the plugin library
 *
 * @return a combination of @ref varjo_PluginType flags
 */
VARJO_API varjo_PluginType varjo_GetPluginType();

/**
 * Provides the API version number supported by the plugin.
 * Use @ref VARJO_PLUGIN_API_VERSION macro to indicate supported API version number.
 *
 * @return Varjo plugin API version number.
 */
VARJO_API uint64_t varjo_GetPluginAPIVersion();

/**
 * Provides plugin name
 *
 * @return Null-terminated C-string containing plugin name
 */
VARJO_API const char* varjo_GetPluginName();

/**
 * Plugin status
 */
typedef int64_t varjo_PluginStatus;
static const varjo_PluginStatus varjo_PluginStatus_OK = 0;  //!< Plugin status OK

/**
 * Provides a string description of a given status code
 *
 * @param pluginStatus - plugin status code
 * @return Null-terminated C-string containing description of a given status code
 */
VARJO_API const char* varjo_GetPluginStatusDescription(varjo_PluginStatus pluginStatus);

/**
 * Settings key enumeration handle
 */
typedef struct varjo_ConfigurationKeyIterator_t* varjo_ConfigurationKeyIterator;

/**
 * @brief varjo_PluginUtilitySession
 */
typedef struct varjo_PluginUtilitySession_t* varjo_PluginUtilitySession;

/**
 * Double precision 4x4 matrix.
 * The matrix usage convention is that values are stored in column-major order.
 */
typedef struct varjo_Matrix4x4 {
    double value[16];
} varjo_Matrix4x4;

/**
 * Bit-field type determining capabilities of IMU hardware
 */
typedef uint64_t varjo_IMUCapability;
static const varjo_IMUCapability varjo_IMUCapability_Acceleration = 0x01;     //!< IMU HW is capable of reporting accelerometer measurements
static const varjo_IMUCapability varjo_IMUCapability_AngularVelocity = 0x02;  //!< IMU HW is capable of reporting gyro measurements
static const varjo_IMUCapability varjo_IMUCapability_Attitude = 0x04;         //!< IMU HW is capable of reporting magnetometer measurements

/**
 * Bit-field type determining which values of IMU data are off-scale.
 * These flags are set when a corresponding sensor reading exceeds a limit of it's
 * measurement range
 */
typedef uint64_t varjo_IMUDataOffscale;
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AccelerationX = 0x001;     //!< Flag indicating the X-axis acceleration measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AccelerationY = 0x002;     //!< Flag indicating the Y-axis acceleration measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AccelerationZ = 0x004;     //!< Flag indicating the Z-axis acceleration measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AngularVelocityX = 0x008;  //!< Flag indicating the X-axis angular velocity measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AngularVelocityY = 0x010;  //!< Flag indicating the Y-axis angular velocity measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AngularVelocityZ = 0x020;  //!< Flag indicating the Z-axis angular velocity measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AttitudeX = 0x040;         //!< Flag indicating the X-axis attitude measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AttitudeY = 0x080;         //!< Flag indicating the Y-axis attitude measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AttitudeZ = 0x100;         //!< Flag indicating the Z-axis attitudey measurement is off-scale

/**
 * Bit-field type determining which values of IMU data are not valid,
 * e.g. in case the IMU HW can not provide corresponding measurement type, or
 * does not have a measurement sample
 */
typedef uint64_t varjo_IMUDataInvalid;
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AccelerationX = 0x001;     //!< Flag indicating the X-axis acceleration measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AccelerationY = 0x002;     //!< Flag indicating the Y-axis acceleration measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AccelerationZ = 0x004;     //!< Flag indicating the Z-axis acceleration measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AngularVelocityX = 0x008;  //!< Flag indicating the X-axis angular velocity measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AngularVelocityY = 0x010;  //!< Flag indicating the Y-axis angular velocity measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AngularVelocityZ = 0x020;  //!< Flag indicating the Z-axis angular velocity measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AttitudeX = 0x040;         //!< Flag indicating the X-axis attitude measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AttitudeY = 0x080;         //!< Flag indicating the Y-axis attitude measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AttitudeZ = 0x100;         //!< Flag indicating the Z-axis attitude measurement is not valid


/**
 * Type identifying IMU component type
 */
typedef int64_t varjo_IMUComponent;
static const varjo_IMUComponent varjo_IMUComponent_Accelerometer = 1;  //!< Constant identifying IMU component accelerometer
static const varjo_IMUComponent varjo_IMUComponent_Gyro = 2;           //!< Constant identifying IMU component gyro
static const varjo_IMUComponent varjo_IMUComponent_Magnetometer = 3;   //!< Constant identifying IMU component magnetometer

/**
 * Structure containing calibration parameters for an IMU device
 */
typedef struct varjo_IMUComponentCalibrationInfo {
    varjo_Matrix3x3 scale;  //!< Calibration scale
    varjo_Vector3D bias;    //!< Calibration bias
} varjo_IMUDeviceCalibrationInfo;

/**
 * Structure describing a sample of IMU data
 */
typedef struct varjo_ImuSample {
    varjo_Nanoseconds timeStamp;          //!< IMU sample timestamp (ns)
    varjo_Vector3D acceleration;          //!< Device acceleration (m/s^2)
    varjo_Vector3D angularVelocity;       //!< Device angular velocity (radians/s)
    varjo_Vector3D attitude;              //!< Device attitude
    varjo_IMUDataOffscale offscaleFlags;  //!< Offscale flags
    varjo_IMUDataInvalid validityFlags;   //!< Validity flags
} varjo_ImuSample;

/**
 * Plugin system state
 */
typedef int64_t varjo_PluginSystemState;
static const varjo_PluginSystemState varjo_PluginSystemState_OK = 0;       //!< Plugin system is functioning correctly
static const varjo_PluginSystemState varjo_PluginSystemState_Warning = 1;  //!< Plugin system is reporting a state requiring user attention e.g. battery low
static const varjo_PluginSystemState varjo_PluginSystemState_CriticalError = 2;  //!< Plugin system is reporting a critical error state, e.g. tracking failure

/**
 * @struct varjo_PluginUtilityAPI
 *
 * Utility functions plugins can call
 */
typedef struct varjo_PluginUtilityAPI {
    /**
     * Returns Varjo product id string
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @returns NULL terminated C-string containing Varjo product id. e.g. "VR-1"
     */
    typedef const char*(VARJO_FNTABLE_CALLTYPE* GetProductId_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetProductId_t. */
    GetProductId_t GetProductId;

    /**
     * Returns hardware capabilities of the IMU
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @returns A combination of @ref varjo_IMUCapability flags describing IMU hardware
     * capabilities.
     */
    typedef varjo_IMUCapability(VARJO_FNTABLE_CALLTYPE* GetIMUCapabilities_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetIMUCapabilities_t. */
    GetIMUCapabilities_t GetIMUCapabilities;

    /**
     * Returns transformation matrix between IMU and the HMD coordinate systems
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @returns A transformation matrix between IMU and HMD coordinate systems.
     */
    typedef varjo_Matrix4x4(VARJO_FNTABLE_CALLTYPE* GetIMUToHMDTransform_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetIMUToHMDTransform_t. */
    GetIMUToHMDTransform_t GetIMUToHMDTransform;

    /**
     * Returns calibration for an IMU component
     *
     * <em>
     * In VR-1, XR-1 and VR-2 series HMDs the IMU is located (from HMD user's perspective):
     * 0.5mm to the right, 13.4mm down and 57.2mm forward from the HMD tracking point.
     *
     * IMU is oriented such that its positive X points right, positive Y points down and
     * positive Z points forward.
     *
     * Calibration parameters are calculated after production and are constant in run time.
     *
     * Calibration can be applied this way: <c>IMUCCalibrated = (IMURaw - Bias) * Scale</c>
     * </em>
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param component - IMU hardare component for which the calibration is requested
     * @returns IMU calibration parameters for a @c component
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* GetIMUComponentCalibration_t)(  //
        varjo_PluginUtilitySession session,                              //
        varjo_IMUComponent component,                                    //
        varjo_IMUComponentCalibrationInfo* calibrationInfo);

    /** @sa GetIMUComponentCalibration_t. */
    GetIMUComponentCalibration_t GetIMUComponentCalibration;

    /**
     * Reads up to @c sampleCount samples of @ref varjo_ImuSample into a @c samples buffer.
     *
     * The @c varjo_ImuSample values obtained via @ref ReadIMUSamples_t are not calibrated.
     * @sa GetIMUComponentCalibration_t
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param samples - Pointer to the buffer where the IMU data samples should be stored
     * @param sampleCount - Number of IMU samples to read
     * @returns Number of IMU samples read
     */
    typedef int32_t(VARJO_FNTABLE_CALLTYPE* ReadIMUSamples_t)(  //
        varjo_PluginUtilitySession session,                     //
        varjo_ImuSample* samples, int32_t sampleCount);

    /** @sa ReadIMUSamples_t. */
    ReadIMUSamples_t ReadIMUSamples;

    /**
     * Atomically applies set configuration
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     */
    typedef varjo_Bool(VARJO_FNTABLE_CALLTYPE* SyncConfiguration_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa SyncConfiguration_t. */
    SyncConfiguration_t SyncConfiguration;

    /**
     * Enumerates configuration keys
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param handle - enumeration handle, pass NULL to start enumeration
     * @return NULL when enumeration is over, otherwise a null-terminated C-string containing
     *         key name.
     */
    typedef const char*(VARJO_FNTABLE_CALLTYPE* EnumConfigurationKeys_t)(  //
        varjo_PluginUtilitySession session,                                //
        varjo_ConfigurationKeyIterator* handle);

    /** @sa EnumConfigurationKeys_t. */
    EnumConfigurationKeys_t EnumConfigurationKeys;

    /**
     * Sets configuration entry @param value for @param key
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* SetConfigurationString_t)(  //
        varjo_PluginUtilitySession session,                          //
        const char* key, const char* value);

    /** @sa SetConfigurationString_t. */
    SetConfigurationString_t SetConfigurationString;

    /**
     * Writes configuration entry @param value for @param key
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @return If @c value is NULL, returns size of buffer necessary for @c value,
     * otherwise length of the value for @c key written to @c value.
     * Returs 0 if the requested configuration key is not found.
     */
    typedef int32_t(VARJO_FNTABLE_CALLTYPE* GetConfigurationString_t)(  //
        varjo_PluginUtilitySession session,                             //
        const char* key, char* value, int32_t valueLength);

    /** @sa GetConfigurationString_t. */
    GetConfigurationString_t GetConfigurationString;


    /**
     * Returns current system time
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @return Current time
     */
    typedef varjo_Nanoseconds(VARJO_FNTABLE_CALLTYPE* GetCurrentSystemTime_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetCurrentSystemTime_t. */
    GetCurrentSystemTime_t GetCurrentSystemTime;

    /**
     * Writes a @c message to the log
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* WriteLogMessage_t)(  //
        varjo_PluginUtilitySession session,                   //
        const char* message);

    /** @sa WriteLogMessage_t. */
    WriteLogMessage_t WriteLogMessage;

    /**
     * Indicates a plugin system @c state to Varjo SW stack.
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param state - a value of type @ref varjo_PluginSystemState indicating plugin system state
     * @param message - optional message containing technical description of given state
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* SetPluginSystemState_t)(  //
        varjo_PluginUtilitySession session,                        //
        varjo_PluginSystemState state,                             //
        const char* message);

    /** @sa SetPluginSystemState_t. */
    SetPluginSystemState_t SetPluginSystemState;

} varjo_PluginUtilityAPI;

/**
 * Plugin session context
 */
typedef struct varjo_PluginContext_t* varjo_PluginContext;

/**
 * Initializes plugin
 *
 * @param pluginUtilityAPI - pointer to structure containing utility function pointers
 * @param pluginUtilitySession - handle to pass to @ref varjo_PluginUtilityAPI functions
 * @param context - pointer to write plugin session context
 * @return @ref varjo_PluginStatus_OK if plugin was initialized successfully, other value of
 *         type @ref varjo_PluginStatus otherwise
 */
VARJO_API varjo_PluginStatus varjo_PluginInit(        //
    const varjo_PluginUtilityAPI* pluginUtilityAPI,   //
    varjo_PluginUtilitySession pluginUtilitySession,  //
    varjo_PluginContext* context);

/**
 * Deinitializes plugin
 *
 * @param context - plugin session context acquired from varjo_PluginInit
 * @return @ref varjo_PluginStatus_OK if plugin was shut down successfully, other value of
 *         type @ref varjo_PluginStatus otherwise
 */
VARJO_API varjo_PluginStatus varjo_PluginShutdown(varjo_PluginContext* context);

// ----------------------------------------------------------------------------
// Tracker Plugin API types

/**
 * Bit-field type determining tracking status
 */
typedef uint64_t varjo_PoseFlags;
static const varjo_PoseFlags varjo_PoseFlags_Ok = 0x1;                    //!< Flag indicating that the tracking is OK
static const varjo_PoseFlags varjo_PoseFlags_TrackingLost = 0x2;          //!< Flag indicating that no tracking data is available
static const varjo_PoseFlags varjo_PoseFlags_TrackingDisconnected = 0x4;  //!< Flag indicating that the tracker is disconnected
static const varjo_PoseFlags varjo_PoseFlags_HasPosition = 0x08;          //!< Flag indicating that the pose has position information
static const varjo_PoseFlags varjo_PoseFlags_HasRotation = 0x10;          //!< Flag indicating that the pose has rotation information
static const varjo_PoseFlags varjo_PoseFlags_HasVelocity = 0x20;          //!< Flag indicating that the pose has velocity information
static const varjo_PoseFlags varjo_PoseFlags_HasAngularVelocity = 0x40;   //!< Flag indicating that the pose has angular velocity information
static const varjo_PoseFlags varjo_PoseFlags_HasAcceleration = 0x80;      //!< Flag indicating that the pose has acceleration information
static const varjo_PoseFlags varjo_PoseFlags_HasConfidence = 0x100;       //!< Flag indicating that the pose has confidence information

/**
 * Double precision quaternion.
 */
typedef struct varjo_Quaternion {
    double w;  //!< Quaternion component w
    double x;  //!< Quaternion component x
    double y;  //!< Quaternion component y
    double z;  //!< Quaternion component z
} varjo_Quaternion;

/**
 * Structure describing tracked device pose.
 *
 * <em>
 * Coordinate system is right-handed, positive X goes right, positive Y goes up
 * and negative Z goes forward from HMD users perspective.
 * Distance unit is meters.
 * </em>
 */
typedef struct varjo_DevicePose {
    varjo_Nanoseconds timeStamp;     //!< Pose timestamp (ns)
    int64_t frameNo;                 //!< Frame number, increases monotonically.
    varjo_Vector3D position;         //!< Device position (m)
    varjo_Quaternion rotation;       //!< Device rotation
    varjo_Vector3D velocity;         //!< Device linear velocity (m/s)
    varjo_Vector3D angularVelocity;  //!< Device angular velocity (radians/s)
    varjo_Vector3D acceleration;     //!< Device acceleration (m/s^2)
    varjo_PoseFlags poseFlags;       //!< Bit field value describing pose
    double confidence;               //!< Tracker confidence
} varjo_DevicePose;

// ----------------------------------------------------------------------------
// Tracker Plugin API

/**
 * Starts the tracker processing
 * @param context - plugin session context acquired from @ref varjo_PluginInit
 * @return @ref varjo_PluginStatus_OK if tracker started successfully, other value of
 *         type @ref varjo_PluginStatus otherwise
 */
VARJO_API varjo_PluginStatus varjo_TrackerStart(varjo_PluginContext context);

/**
 * Stops the tracker processing
 * @param context - plugin session context acquired from @ref varjo_PluginInit
 * @return @ref varjo_PluginStatus_OK if tracker stopped successfully, other value of
 *         type @ref varjo_PluginStatus otherwise
 */
VARJO_API varjo_PluginStatus varjo_TrackerStop(varjo_PluginContext context);

/**
 * @brief Queries tracker for the HMD pose
 *
 * <em>
 * HMD's tracking point is expected to be located in the middle - between the
 * centers of eye-piece lenses, and the on plane of the lens surface closest to the user eyes.
 * </em>
 *
 * @image html HMD_Angled.png width=30%
 * @image html HMD_Side.png width=30%
 * @image html HMD_Top.png width=30%
 *
 * @param context - plugin session context acquired from @ref varjo_PluginInit
 * @param timeStamp - timestamp in nanoseconds for which tracking prediction is requested
 * @param pose - Pointer to a structure receiving tracking
 * @return @ref varjo_PluginStatus_OK if HMD pose was queried successfully, other value of
 *         type @ref varjo_PluginStatus otherwise
 */
VARJO_API varjo_PluginStatus varjo_GetHMDPose(  //
    varjo_PluginContext context,                //
    varjo_Nanoseconds timeStamp,                //
    varjo_DevicePose* pose);

#if defined __cplusplus
}  // extern "C"
#endif  // __cplusplus

#endif  // VARJO_PLUGIN_H
